@@ -1,3 +1,4 @@
+// -*- Mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*-
 // Copyright (c) 2006, Google Inc.
 // All rights reserved.
 // 
@@ -123,6 +124,9 @@ const char* GetenvBeforeMain(const char* name) {
   if (__environ) {            // can exist but be NULL, if statically linked
     const int namelen = strlen(name);
     for (char** p = __environ; *p; p++) {
+      if (strlen(*p) < namelen) {
+        continue;
+      }
       if (!memcmp(*p, name, namelen) && (*p)[namelen] == '=')  // it's a match
         return *p + namelen+1;                                 // point after =
     }
@@ -166,6 +170,12 @@ const char* GetenvBeforeMain(const char* name) {
   return NULL;                   // env var never found
 }
 
+extern "C" {
+  const char* TCMallocGetenvSafe(const char* name) {
+    return GetenvBeforeMain(name);
+  }
+}
+
 // This takes as an argument an environment-variable name (like
 // CPUPROFILE) whose value is supposed to be a file-path, and sets
 // path to that path, and returns true.  If the env var doesn't exist,
