@@ -226,7 +226,7 @@
 #define PACKAGE_NAME "gperftools"
 
 /* Define to the full name and version of this package. */
-#define PACKAGE_STRING "gperftools 2.2.1"
+#define PACKAGE_STRING "gperftools 2.4"
 
 /* Define to the one symbol short name of this package. */
 #define PACKAGE_TARNAME "gperftools"
@@ -235,7 +235,7 @@
 #undef PACKAGE_URL
 
 /* Define to the version of this package. */
-#define PACKAGE_VERSION "2.2.1"
+#define PACKAGE_VERSION "2.4"
 
 /* How to access the PC from a struct ucontext */
 #undef PC_FROM_UCONTEXT
